_default_config = dict(language='english')

class Config(dict):
    
    def __init__(self, **kwargs):
        for key in kwargs:
            self[key] = kwargs[key]
            
    def update(self, **kwargs):
        for key in kwargs:
            self[key] = kwargs[key]
            
config = Config(**_default_config)


"""
"This" is my example-script
===========================

This example doesn't do much, it just makes a simple plot
"""

# %
# Test sub section
# -------------------------
#
# In the built documentation, it will be rendered as reST. All reST lines
# must begin with '# ' (note the space) including underlines below section
# headers.

import plotly
import copy
import os

import numpy as np

import plotly.graph_objects as go
import plotly.graph_objs._figure


def plot_topo_3D(
    z: np.ndarray,
    x: np.ndarray,
    y: np.ndarray,
    ndv: int | float = -9999,
    vert_exag: float = 1,
    light_source: tuple[int, int] = None,
    add_walls: bool = True,
    add_floor: bool = True,
    saving_path: str = None,
    auto_open: bool = False
) -> plotly.graph_objs._figure.Figure:
    """
    Plot 3D topography with hillshading.

    Parameters
    ----------
    z : np.ndarray
        2D array of elevation values.
    x : np.ndarray
        1D array of X coordinates.
    y : np.ndarray
        1D array of Y coordinates.
    ndv : int or float, optional
        No-data value in the elevation array. The default is -9999.
    vert_exag : float, optional
        Vertical exaggeration factor. The default is 1.
    light_source : tuple[int], optional
        Light source direction for shading (azimuth, degrees from North and altitude, degrees from horizon). The default is None.
    add_walls : bool, optional
        If True, add walls under the figure. The default is True
    add_floor : bool, optional
        If True, add floor under the figure. The default is True
    saving_path : str, optional
        Folder path to save the html file. If None file saved in source folder. The default is None.
    auto_open : bool, optional
        If True, open the result. The default is False.

    Returns
    -------
    None
    """
    corrz = np.copy(z)
    z_base = np.nanmin(corrz)

    corrz[corrz == ndv] = np.nan

    if light_source is not None:
        az = np.radians(light_source[0])
        alt = np.radians(light_source[1])
    else:
        az = np.radians(90)
        alt = np.radians(45)
    

    fig = go.Figure(data=[
        go.Surface(
            z=corrz*vert_exag,
            x=x,
            y=y,
            surfacecolor=np.zeros_like(corrz),
            colorscale=[[0, 'lightgray'], [1, 'lightgray']],
            cmin=np.nanmin(z),
            cmax=np.nanmax(z),
            showscale=False,
            lighting=dict(ambient=0.5, diffuse=0.8, specular=0.5,
                          roughness=0.5, fresnel=0.2),
            lightposition=dict(x=float(np.cos(alt) * np.cos(az)),
                               y=float(np.cos(alt) * np.sin(az)), z=float(np.sin(alt)))
        )
    ])

    if add_walls:
        fig.add_trace(go.Surface(
            z=np.vstack([corrz[0], np.full_like(z[0], z_base)]),
            x=np.tile(x, (2, 1)),
            y=np.full((2, len(x)), y[0]),
            showscale=False, colorscale=[[0, 'gray'], [1, 'gray']], opacity=0.8
        ))

        fig.add_trace(go.Surface(
            z=np.vstack([corrz[-1], np.full_like(z[-1], z_base)]),
            x=np.tile(x, (2, 1)),
            y=np.full((2, len(x)), y[-1]),
            showscale=False, colorscale=[[0, 'gray'], [1, 'gray']], opacity=0.8
        ))

        fig.add_trace(go.Surface(
            z=np.column_stack([corrz[:, 0], np.full_like(z[:, 0], z_base)]),
            x=np.full((len(y), 2), x[0]),
            y=np.tile(y, (2, 1)).T,
            showscale=False, colorscale=[[0, 'gray'], [1, 'gray']], opacity=0.8
        ))

        fig.add_trace(go.Surface(
            z=np.column_stack([corrz[:, -1], np.full_like(z[:, -1], z_base)]),
            x=np.full((len(y), 2), x[-1]),
            y=np.tile(y, (2, 1)).T,
            showscale=False, colorscale=[[0, 'gray'], [1, 'gray']], opacity=0.8
        ))

    if add_floor:
        fig.add_trace(go.Surface(
            z=np.full_like(corrz, z_base),
            x=x,
            y=y,
            showscale=False, colorscale=[[0, 'gray'], [1, 'gray']], opacity=0.8
        ))

    fig.update_layout(
        title='Topographic Surface',
        scene=dict(
            xaxis_title='X (m)',
            yaxis_title='Y (m)',
            zaxis_title='Z (m)',
            aspectmode='auto'
        )
    )

    if saving_path is not None:
        fig.write_html(os.path.join(
            saving_path, "topography.html"), auto_open=auto_open)
    
    return fig


def plot_imshow_3D(
    x: np.ndarray,
    y: np.ndarray,
    data: np.ndarray,
    cmap: str = None,
    minval: int | float = None,
    maxval: int | float = None,
    minval_abs: float = None,
    vert_exag: float = 1.,
    saving_path: str = None,
    auto_open: bool = False
) -> plotly.graph_objs._figure.Figure:
    """3D imshow data

    jfdqsjf

    Parameters
    ----------
    x : np.ndarray
        1D array of X coordinates.
    y : np.ndarray
        1D array of Y coordinates.
    data : np.ndarray
        2D array of data values.
    cmap : str, optional
        Name of the colormap to use. If None, use "Hot" or "Rdbu. The default is None.
    minval : int or float, optional
        Minimum value to display. The default is None.
    maxval : int or float, optional
        Maximum value to display. The default is None.
    minval_abs : float, optional
        If set, masks values with absolute magnitude below this threshold. The default is None.
    vert_exag : float, optional
        Vertical exaggeration factor. The default is 1.
    saving_path : str, optional
        Folder path to save the html file. If None file saved in source folder. The default is None.
    auto_open : bool, optional
        If True, open the result. The default is False.

    Returns
    -------
    _type_
        _description_
    """
    f = copy.copy(data)

    if minval_abs is not None:
        f[np.abs(f) <= minval_abs] = np.nan
    else:
        f[f == 0] = np.nan

    if minval is None:
        minval = f.min()

    if maxval is None:
        maxval = f.min()

    if cmap is None:
        if (maxval*minval) <= 0:
            cmap = 'Rdbu'
        else:
            cmap = 'Hot'

    fig = go.Figure(data=[
        go.Surface(
            z=f*vert_exag,
            x=x,
            y=y,
            colorscale=cmap,
            cmin=minval,
            cmax=maxval,
            showscale=True
        )
    ])

    fig.update_layout(
        title='Data Surface',
        scene=dict(
            xaxis_title='X (m)',
            yaxis_title='Y (m)',
            zaxis_title='Z (m)',
            aspectmode='auto'
        )
    )

    if saving_path is not None:
        fig.write_html(os.path.join(
            saving_path, "data.html"), auto_open=auto_open)
    
    return fig


def plot_data_on_topo_3D(
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray,
    data: np.ndarray,
    ndv: int | float = -9999,
    vert_exag: float = 1,
    light_source: tuple[int, int] = None,
    cmap: str = None,
    minval: int | float = None,
    maxval: int | float = None,
    minval_abs: float = None,
    add_walls: bool = True,
    add_floor: bool = True,
    saving_path: str = None,
    auto_open: bool = False
) -> plotly.graph_objs._figure.Figure:
    """
    Plot 3D array data on topo.

    Parameters
    ----------
    x : np.ndarray
        1D array of X coordinates.
    y : np.ndarray
        1D array of Y coordinates.
    z : np.ndarray  
        2D array of elevation values.
    data : np.ndarray
        2D array of data values.
    ndv : int or float, optional
        No-data value in the elevation array. The default is -9999.
    light_source : tuple[int], optional
        Light source direction for shading (azimuth, degrees from North and altitude, degrees from horizon). The default is None.
    cmap : str, optional
        Name of the colormap to use. If None, use "Hot" or "Rdbu. The default is None.
    minval : int or float, optional
        Minimum value to display. The default is None.
    maxval : int or float, optional
        Maximum value to display. The default is None.
    minval_abs : float, optional
        If set, masks values with absolute magnitude below this threshold. The default is None.
    add_walls : bool, optional
        If True, add walls under the figure. The default is True
    add_floor : bool, optional
        If True, add floor under the figure. The default is True
    saving_path : str, optional
        Folder path to save the html file. If None file saved in source folder. The default is None.
    auto_open : bool, optional
        If True, open the result. The default is False.

    Returns
    -------
    None
    """
    corrz = np.copy(z)
    z_base = np.nanmin(corrz)

    corrz[corrz == ndv] = np.nan

    f = copy.copy(data)

    if minval_abs is not None:
        f[np.abs(f) <= minval_abs] = np.nan
    else:
        f[f == 0] = np.nan

    if minval is None:
        minval = f.min()

    if maxval is None:
        maxval = f.min()

    if cmap is None:
        if (maxval*minval) <= 0:
            cmap = 'Rdbu'
        else:
            cmap = 'Hot'
    
    if light_source is not None:
        az = np.radians(light_source[0])
        alt = np.radians(light_source[1])
    else:
        az = np.radians(90)
        alt = np.radians(45)

    new_data = np.where(~np.isnan(f), f + z + 1, np.nan)

    fig = go.Figure(data=[
        go.Surface(
            z=corrz,
            x=x,
            y=y,
            surfacecolor=np.zeros_like(corrz),
            colorscale=[[0, 'lightgray'], [1, 'lightgray']],
            cmin=np.nanmin(z),
            cmax=np.nanmax(z),
            showscale=False,
            lighting=dict(ambient=0.5, diffuse=0.8, specular=0.5,
                          roughness=0.5, fresnel=0.2),
            lightposition=dict(x=float(np.cos(alt) * np.cos(az)),
                               y=float(np.cos(alt) * np.sin(az)), z=float(np.sin(alt)))
        ),
        go.Surface(
            z=new_data,
            x=x,
            y=y,
            surfacecolor=f,
            colorscale=cmap,
            cmin=minval,
            cmax=maxval,
            showscale=True,
            colorbar=dict(title="Data thickness", thickness=15)
        )
    ])

    if add_walls:
        fig.add_trace(go.Surface(
            z=np.vstack([corrz[0], np.full_like(z[0], z_base)]),
            x=np.tile(x, (2, 1)),
            y=np.full((2, len(x)), y[0]),
            showscale=False, colorscale=[[0, 'gray'], [1, 'gray']], opacity=0.8
        ))

        fig.add_trace(go.Surface(
            z=np.vstack([corrz[-1], np.full_like(z[-1], z_base)]),
            x=np.tile(x, (2, 1)),
            y=np.full((2, len(x)), y[-1]),
            showscale=False, colorscale=[[0, 'gray'], [1, 'gray']], opacity=0.8
        ))

        fig.add_trace(go.Surface(
            z=np.column_stack([corrz[:, 0], np.full_like(z[:, 0], z_base)]),
            x=np.full((len(y), 2), x[0]),
            y=np.tile(y, (2, 1)).T,
            showscale=False, colorscale=[[0, 'gray'], [1, 'gray']], opacity=0.8
        ))

        fig.add_trace(go.Surface(
            z=np.column_stack([corrz[:, -1], np.full_like(z[:, -1], z_base)]),
            x=np.full((len(y), 2), x[-1]),
            y=np.tile(y, (2, 1)).T,
            showscale=False, colorscale=[[0, 'gray'], [1, 'gray']], opacity=0.8
        ))

    if add_floor:
        fig.add_trace(go.Surface(
            z=np.full_like(corrz, z_base),
            x=x,
            y=y,
            showscale=False, colorscale=[[0, 'gray'], [1, 'gray']], opacity=0.8
        ))

    fig.update_layout(
        title='Data on Topography',
        scene=dict(
            xaxis_title='X (m)',
            yaxis_title='Y (m)',
            zaxis_title='Z (m)',
            aspectmode='auto'
        )
    )

    if saving_path is not None:
        fig.write_html(os.path.join(
            saving_path, "data_on_topography.html"), auto_open=auto_open)

    return fig



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  2 16:16:39 2021

@author: peruzzetto
"""

import matplotlib

import numpy as np
import matplotlib.colors as mcolors

BOLD_CONTOURS_INTV = [
    0.1,
    0.2,
    0.5,
    1,
    2.0,
    5,
    10,
    20,
    50,
    100,
    200,
    500,
    1000,
]
NB_THIN_CONTOURS = 10
NB_BOLD_CONTOURS = 3


def centered_map(
    cmap: str,
    vmin: float,
    vmax: float,
    ncolors: int = 256
) -> matplotlib.colors.LinearSegmentedColormap:
    """
    Create centered colormap.

    Parameters
    ----------
    cmap : str
        Curent colormap of the plot.
    vmin : float
        Minimum value of the dataset.
    vmax : float
        Maximal value of the dataset.
    ncolors : int, optional
        Total number of colors contained in the new colormap. The default is 256.

    Returns
    -------
    new_map : matplotlib.colors.LinearSegmentedColormap
        New colormap objects centered in 0.
    """
    p = vmax / (vmax - vmin)
    npos = int(ncolors * p)
    if isinstance(cmap, list):
        cmap_tmp = mcolors.LinearSegmentedColormap.from_list("tmp", cmap)
    else:
        cmap_tmp = matplotlib.colormaps[cmap]
    colors1 = cmap_tmp(np.linspace(0.0, 1, npos * 2))
    colors2 = cmap_tmp(np.linspace(0.0, 1, (ncolors - npos) * 2))
    colors = np.concatenate(
        (colors2[: ncolors - npos, :], colors1[npos:, :]), axis=0
    )
    # colors[ncolors-npos-1,:]=np.ones((1,4))
    # colors[ncolors-npos,:]=np.ones((1,4))
    new_map = mcolors.LinearSegmentedColormap.from_list("my_colormap", colors)

    return new_map


def get_contour_intervals(
    zmin: float,
    zmax: float,
    nb_bold_contours: int = None,
    nb_thin_contours: int = None
) -> tuple[float, float]:
    """
    Provides the intervals for thin and bold contour lines

    Parameters
    ----------
    zmin : float
        Minimum altitude of the topography.
    zmax : float
        Maximum altitude of the topography.
    nb_bold_contours : int, optional
        Number of bold line. The default is None.
    nb_thin_contours : int, optional
        Number of thin line. The default is None.

    Returns
    -------
    bold_intv : float
        Interval between bold contour lines.

    thin_intv : float
        Interval between thin contour lines.
    """
    if nb_thin_contours is None:
        nb_thin_contours = NB_THIN_CONTOURS
    if nb_bold_contours is None:
        nb_bold_contours = NB_BOLD_CONTOURS

    intv = (zmax - zmin) / nb_bold_contours
    i = np.argmin(np.abs(np.array(BOLD_CONTOURS_INTV) - intv))

    bold_intv = BOLD_CONTOURS_INTV[i]
    if BOLD_CONTOURS_INTV[i] != BOLD_CONTOURS_INTV[0]:
        if bold_intv - intv > 0:
            bold_intv = BOLD_CONTOURS_INTV[i - 1]

    if nb_thin_contours is None:
        thin_intv = bold_intv / NB_THIN_CONTOURS
        if (zmax - zmin) / bold_intv > 5:
            thin_intv = thin_intv * 2
    else:
        thin_intv = bold_intv / nb_thin_contours

    return bold_intv, thin_intv


def auto_uniform_grey(
    z: np.ndarray,
    vert_exag: float,
    azdeg: int = 315,
    altdeg: int = 45,
    dx: float = 1.,
    dy: float = 1.,
    std_threshold: float = 0.01
) -> float:
    """
    Detect if shading must be applied to topography or not (uniform grey). The
    criterion in colors.LightSource.hillshade is the difference between min
    and max illumination, and seems to restrictive.

    Parameters
    ----------
    z : np.ndarray
        Altitude of each point of the topography.

    vert_exag : float
        Vertical exaggeration factor for hillshading.

    azdeg : int, optional
        Azimuth angle for light source (degrees from North). The default is 315. 

    altdeg : int, optional
        Altitude angle for light source (degrees above horizon). The default is 45

    dx : float, optional
        Cell size of the x axis. The default is 1.0.

    dy : float, optional
        Cell size of the y axis. The default is 1.0.

    std_treshold: float, optional
        Relief detection threshold. The default is 0.5.

    Returns
    -------
    float | None
        If no shading applied return None, else return 0.5.

    """
    # Get topography normal direction
    e_dy, e_dx = np.gradient(vert_exag * z, dy, dx)
    normal = np.empty(z.shape + (3,)).view(type(z))
    normal[..., 0] = -e_dx
    normal[..., 1] = -e_dy
    normal[..., 2] = 1
    sum_sq = 0
    for i in range(normal.shape[-1]):
        sum_sq += normal[..., i, np.newaxis] ** 2
    normal /= np.sqrt(sum_sq)

    # Light source direction
    az = np.radians(90 - azdeg)
    alt = np.radians(altdeg)
    direction = np.array(
        [np.cos(az) * np.cos(alt), np.sin(az) * np.cos(alt), np.sin(alt)]
    )

    # Compute intensity (equivalent to LightSource hillshade, whithour rescaling)
    intensity = normal.dot(direction)
    std = np.nanstd(intensity)

    if std > std_threshold:
        return None
    else:
        return 0.5


def colorbar(
    mappable,
    ax=None,
    cax=None,
    size: str = "5%",
    pad: float = 0.1,
    position: str = "right",
    **kwargs: dict
) -> matplotlib.colorbar.Colorbar:
    """
    Create nice colorbar matching height/width of axe.

    Parameters
    ----------
    mappable : matplotlib.cm.ScalarMappable
        Mappable object used to generate the colorbar.
    ax : matplotlib.axes.Axes, optional
        The axis associated with the colorbar. The default is None.
    cax : matplotlib.axes.Axes, optional
        Specific axis to draw the colorbar into. The default is None.
    size : str, optional
        Width or height of the colorbar relative to the parent axis. The default is "5%".
    pad : float, optional
        Padding between the parent axis and the colorbar, as a fraction of the parent axis size. The default is 0.1.
    position : str, optional
        Position of the colorbar relative to the parent axis. Options: 'right', 'left', 'top', 'bottom'. The default is "right".
    **kwargs : dict
        Additional keyword arguments passed to "fig.colorbar()".

    Returns
    -------
    matplotlib.colorbar.Colorbar
        The created Colorbar object.
    """
    if ax is None:
        ax = mappable.axes
    fig = ax.figure
    if position in ["left", "right"]:
        orientation = "vertical"
    else:
        orientation = "horizontal"

    # if cax is None:
    #     # divider = ax.get_axes_locator()
    #     # if divider is None:
    #     divider = make_axes_locatable(ax)
    #     cax = divider.append_axes(position, size=size, pad=pad)

    cc = fig.colorbar(mappable, cax=cax, orientation=orientation, **kwargs)

    if position == "top":
        cax.xaxis.tick_top()
        cax.xaxis.set_label_position("top")
    if position == "left":
        cax.yaxis.tick_left()
        cax.xaxis.set_label_position("left")
    return cc


def read_tiff(
    file
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Read tiff file to numpy ndarray.

    Parameters
    ----------
    file : str
        Path to the tiff file.

    Returns
    -------
    tuple[np.ndarray, np.ndarray, np.ndarray]
        x, y, z values of the tiff file.
    """
    import rasterio

    with rasterio.open(file, "r") as src:
        dem = src.read(1)
        ny, nx = dem.shape
        x = np.linspace(src.bounds.left, src.bounds.right, nx)
        y = np.linspace(src.bounds.bottom, src.bounds.top, ny)
    return x, y, dem


def read_ascii(
    file
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Read ascii grid file to numpy ndarray.

    Parameters
    ----------
    file : str
        Path to the ascii file.

    Returns
    -------
    tuple[np.ndarray, np.ndarray, np.ndarray]
        x, y, z values of the ascii file.
    """
    dem = np.loadtxt(file, skiprows=6)
    grid = {}
    with open(file, "r") as fid:
        for i in range(6):
            tmp = fid.readline().split()
            grid[tmp[0]] = float(tmp[1])
    try:
        x0 = grid["xllcenter"]
        y0 = grid["yllcenter"]
    except KeyError:
        x0 = grid["xllcorner"]
        y0 = grid["yllcorner"]
    nx = int(grid["ncols"])
    ny = int(grid["nrows"])
    dx = dy = grid["cellsize"]
    x = np.linspace(x0, x0 + (nx - 1) * dx, nx)
    y = np.linspace(y0, y0 + (ny - 1) * dy, ny)

    return x, y, dem


def read_raster(
    file
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Extract x, y, z values of a tif/ascii file.

    Parameters
    ----------
    file : str
        Path to the file.

    Returns
    -------
    tuple[np.ndarray, np.ndarray, np.ndarray]
        x, y, z values of the file.
    """
    if file.endswith(".asc") or file.endswith(".txt"):
        return read_ascii(file)
    elif file.endswith(".tif") or file.endswith(".tif"):
        return read_tiff(file)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  2 16:16:39 2021

@author: peruzzetto
"""

import pytest
import os

import numpy as np

from pytopomap.plot_3d import plot_topo_3D, plot_imshow_3D, plot_data_on_topo_3D

temporary_folder = r".\\temp_3d"
os.mkdir(temporary_folder)

x = np.linspace(0, 5, 100)
y = np.linspace(0, 5, 100)


def z_hill(x, y):
    z = np.zeros((len(x), len(y)))
    for i in range(len(x)):
        for j in range(len(y)):
            z[i][j] = 1 - ((x[i] - 2.5)**2)/2 - ((y[j] - 2.5)**2)/4
    return z


def z_complex_topo(x, y):
    z = np.zeros((len(x), len(y)))
    for i in range(len(x)):
        for j in range(len(y)):
            z[i][j] = (np.sin(x[i]) * np.cos(y[j]) + 0.5 * np.sin(3 * x[i] + 1) *
                       np.cos(2 * y[j] + 1) + 0.25 * np.sin(5 * x[i] + 2) * np.cos(4 * y[j] + 2))
    return z


@pytest.mark.parametrize("elevation", [
    np.zeros((100, 100)),
    z_hill(x, y),
    z_complex_topo(x, y),
])
def test_topo(elevation):
    plot_topo_3D(elevation, x, y, saving_path=temporary_folder)
    assert len(os.listdir(temporary_folder)) == 1
    clean_file()


def data_hill_1(x, y):
    data = np.zeros((len(x), len(y)))
    for i in range(len(x)):
        for j in range(len(y)):
            data[i][j] = 1 - ((x[i] - 2.5)**2)/3 - ((y[j] - 2.5)**2)/0.5
    return data


def test_imshow():
    plot_imshow_3D(x, y, data_hill_1(x, y), saving_path=temporary_folder)
    assert len(os.listdir(temporary_folder)) == 1
    clean_file()


def test_data_topo():
    plot_data_on_topo_3D(x, y, z_hill(x, y), data_hill_1(x, y), saving_path=temporary_folder)
    assert len(os.listdir(temporary_folder)) == 1
    clean_file()
    remove_folder()


def clean_file():
    if os.path.exists(os.path.join(temporary_folder, "topography.html")):
        os.remove(os.path.join(temporary_folder, "topography.html"))
    if os.path.exists(os.path.join(temporary_folder, "data.html")):
        os.remove(os.path.join(temporary_folder, "data.html"))
    if os.path.exists(os.path.join(temporary_folder, "data_on_topography.html")):
        os.remove(os.path.join(temporary_folder, "data_on_topography.html"))

def remove_folder():
    if len(os.listdir(temporary_folder)) == 0:
        os.rmdir(temporary_folder)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  2 16:16:39 2021

@author: peruzzetto
"""

import matplotlib
import pytest
import os

import numpy as np

from pytopomap.plot import plot_topo, plot_imshow, plot_data_on_topo, plot_maps

temporary_folder = r".\\temp"
os.mkdir(temporary_folder)

x = np.linspace(0, 5, 100)
y = np.linspace(0, 5, 100)


def z_hill(x, y):
    z = np.zeros((len(x), len(y)))
    for i in range(len(x)):
        for j in range(len(y)):
            z[i][j] = 1 - ((x[i] - 2.5)**2)/2 - ((y[j] - 2.5)**2)/4
    return z


def z_complex_topo(x, y):
    z = np.zeros((len(x), len(y)))
    for i in range(len(x)):
        for j in range(len(y)):
            z[i][j] = (np.sin(x[i]) * np.cos(y[j]) + 0.5 * np.sin(3 * x[i] + 1) *
                       np.cos(2 * y[j] + 1) + 0.25 * np.sin(5 * x[i] + 2) * np.cos(4 * y[j] + 2))
    return z


@pytest.mark.parametrize("elevation", [
    np.zeros((100, 100)),
    z_hill(x, y),
    z_complex_topo(x, y),
])
def test_topo(elevation):
    assert isinstance(plot_topo(elevation, x, y), matplotlib.axes._axes.Axes)


def data_hill_1(x, y):
    data = np.zeros((len(x), len(y)))
    for i in range(len(x)):
        for j in range(len(y)):
            data[i][j] = 1 - ((x[i] - 2.5)**2)/3 - ((y[j] - 2.5)**2)/0.5
    return data


def data_hill_2(x, y):
    data = np.zeros((len(x), len(y)))
    for i in range(len(x)):
        for j in range(len(y)):
            data[i][j] = 1 - ((x[i] - 2.5)**2)/3 - ((y[j] - 2.5)**2)/0.5
    return data


def test_topo_with_param():
    assert isinstance(plot_topo(z_hill(x, y), x, y, contour_step=2, nlevels=2, level_min=-1.5, step_contour_bold=0.5, label_contour=False,
                      vert_exag=2, figsize=(12.1, 5.9), sea_level=-2.5, sea_color='blue', alpha=0.5, azdeg=20), matplotlib.axes._axes.Axes)


def test_imshow():
    assert isinstance(plot_imshow(x, y, data_hill_1(x, y)),
                      matplotlib.axes._axes.Axes)


def test_imshow_with_param():
    assert isinstance(plot_imshow(x, y, data_hill_1(x, y), figsize=(12.1, 4.6), cmap="viridis", minval=0.1, maxval=0.9,
                      alpha=0.9, cmap_intervals=[0.2, 0.3, 0.4, 0.5, 0.6], plot_colorbar=False), matplotlib.axes._axes.Axes)


def test_data_on_topo():
    assert isinstance(plot_data_on_topo(x, y, np.zeros(
        (100, 100)), data_hill_1(x, y)), matplotlib.axes._axes.Axes)


def test_data_on_topo_with_param():
    assert isinstance(plot_data_on_topo(x, y, z_hill(x, y), data_hill_1(x, y), figsize=(10, 10), cmap='viridis', minval=0.1, maxval=0.9, alpha=0.9, cmap_intervals=[
                      0.2, 0.3, 0.4, 0.5, 0.6], plot_colorbar=False, topo_kwargs={"sea_level": -2.5, "sea_color": 'blue'}), matplotlib.axes._axes.Axes)


def test_maps():
    plot_maps(x, y, z_hill(x, y), np.stack((data_hill_1(x, y), data_hill_2(
        x, y)), axis=-1), np.array([0.1, 0.2]), 'Test', temporary_folder)
    assert len(os.listdir(temporary_folder)) == 2
    os.remove(os.path.join(temporary_folder, "Test_0000.png"))
    os.remove(os.path.join(temporary_folder, "Test_0001.png"))
    os.rmdir(temporary_folder)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  2 16:16:39 2021

@author: peruzzetto
"""

import matplotlib
import copy
import os

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors

from pytopomap.tools import centered_map, get_contour_intervals, auto_uniform_grey
# from pytopomap.tools import colorbar


def plot_topo(
    z: np.ndarray,
    x: np.ndarray,
    y: np.ndarray,
    contour_step: float = None,
    nlevels: int = None,
    level_min: float = None,
    step_contour_bold: float | str = "auto",
    contour_labels_properties: dict = None,
    label_contour: bool = True,
    contour_label_effect: list = None,
    axe: matplotlib.axes._axes.Axes = None,
    vert_exag: float = 1.,
    ndv: int | float = -9999,
    uniform_grey: str = "auto",
    contours_prop: dict = None,
    contours_bold_prop: dict = None,
    figsize: tuple[float] = None,
    interpolation: str = None,
    sea_level: int | float = 0,
    sea_color: str = None,
    alpha: float = 1.,
    azdeg: int | float = 315,
    altdeg: int | float = 45,
    ) -> matplotlib.axes._axes.Axes: 
    """
    Plot topography with hillshading.

    Parameters
    ----------
    z : np.ndarray
        2D array of elevation values.
    x : np.ndarray
        1D array of X coordinates.
    y : np.ndarray
        1D array of Y coordinates.
    contour_step : float, optional
        Interval between thin contour lines. Automatically determined if None. The default is None.
    nlevels : TYPE, optional
        Number of contour levels to draw if "contour_step" is not set. The default is None.
    level_min : float, optional
        Minimum contour level. Automatically calculated if None. The default is None.
    step_contour_bold : float or str, optional
        Interval between bold contour lines. If "auto", computed from elevation range. The default is "auto".
    contour_labels_properties : dict, optional
        Properties passed to "ax.clabel()" for contour labels. The default is None.
    label_contour : bool, optional
        If True, add labels to bold contour lines. The default is True.
    contour_label_effect : list, optional
        List of matplotlib.patheffects to apply to contour labels. The default is None.
    axe : matplotlib.axes._axes.Axes, optional
        Existing axes object to draw the plot on. The default is None.
    vert_exag : float, optional
        Vertical exaggeration factor for hillshading. The default is 1.
    ndv : int or float, optional
        No-data value in the elevation array. The default is -9999.
    uniform_grey : str, optional
        If "auto", use auto_uniform_grey to compute the hillshading. The default is "auto".
    contours_prop : dict, optional
        Properties for thin contour lines. The default is "auto".
    contours_bold_prop : dict, optional
        Properties for bold contour lines. The default is "auto".
    figsize : tuple[float], optional
        Figure size (width, height) if a new figure is created. The default is None.
    interpolation : str, optional
        Interpolation method for "imshow". The default is None.
    sea_level : int or float, optional
        Sea level threshold for optional sea overlay. The default is 0.
    sea_color : str, optional
        Color used to fill areas below sea level. If None, no sea mask is applied. The default is None.
    alpha : float, optional
        Transparency for the hillshading layer. The default is 1.
    azdeg : int or float, optional
        Azimuth angle for light source (degrees from North). The default is 315.
    altdeg : int or float, optional
        Altitude angle for light source (degrees above horizon). The default is 45.

    Returns
    -------
    axe : matplotlib.axes._axes.Axes
        Axes object containing the plotted topography.
    """
    dx = x[1] - x[0]
    dy = y[1] - y[0]
    im_extent = [x[0] - dx / 2, x[-1] + dx / 2, y[0] - dy / 2, y[-1] + dy / 2]
    ls = mcolors.LightSource(azdeg=azdeg, altdeg=altdeg)

    auto_bold_intv = None

    tmpz = np.copy(z)
    tmpz[tmpz == ndv] = np.nan

    if nlevels is None and contour_step is None:
        auto_bold_intv, contour_step = get_contour_intervals(
            np.nanmin(tmpz), np.nanmax(tmpz)
        )

    if level_min is None:
        if contour_step is not None:
            level_min = np.ceil(np.nanmin(tmpz) / contour_step) * contour_step
        else:
            level_min = np.nanmin(tmpz)
    if contour_step is not None:
        levels = np.arange(level_min, np.nanmax(tmpz), contour_step)
    else:
        levels = np.linspace(level_min, np.nanmax(tmpz), nlevels)

    if axe is None:
        fig, axe = plt.subplots(1, 1, figsize=figsize, layout="constrained")

    axe.set_ylabel("Y (m)")
    axe.set_xlabel("X (m)")
    axe.set_aspect("equal")

    if uniform_grey == "auto":
        uniform_grey = auto_uniform_grey(
            z,
            vert_exag,
            azdeg=azdeg,
            altdeg=altdeg,
            dx=dx,
            dy=dx,
        )

    z_shade = np.copy(z)
    z_shade[np.isnan(z_shade)] = ndv
    if uniform_grey is None:
        shaded_topo = ls.hillshade(
            z_shade, vert_exag=vert_exag, dx=dx, dy=dy, fraction=1
        )
    else:
        shaded_topo = np.ones(z.shape) * uniform_grey
    shaded_topo[z == ndv] = np.nan
    shaded_topo[np.isnan(z)] = np.nan
    axe.imshow(
        shaded_topo,
        cmap="gray",
        extent=im_extent,
        interpolation=interpolation,
        alpha=alpha,
        vmin=0,
        vmax=1,
    )

    if contours_prop is None:
        contours_prop = dict(alpha=0.5, colors="k", linewidths=0.5)
    axe.contour(
        x,
        y,
        np.flip(tmpz, axis=0),
        extent=im_extent,
        levels=levels,
        **contours_prop
    )

    if contours_bold_prop is None:
        contours_bold_prop = dict(alpha=0.8, colors="k", linewidths=0.8)

    if step_contour_bold == "auto":
        if auto_bold_intv is None:
            auto_bold_intv, _ = get_contour_intervals(
                np.nanmin(tmpz), np.nanmax(tmpz)
            )
        step_contour_bold = auto_bold_intv

    if step_contour_bold > 0:
        lmin = np.ceil(np.nanmin(tmpz) / step_contour_bold) * step_contour_bold
        if lmin < level_min:
            lmin = lmin + step_contour_bold
        levels = np.arange(lmin, np.nanmax(tmpz), step_contour_bold)
        cs = axe.contour(
            x,
            y,
            np.flip(tmpz, axis=0),
            extent=im_extent,
            levels=levels,
            **contours_bold_prop
        )
        if label_contour:
            if contour_labels_properties is None:
                contour_labels_properties = {}
            clbls = axe.clabel(cs, **contour_labels_properties)
            if contour_label_effect is not None:
                plt.setp(clbls, path_effects=contour_label_effect)

    if sea_color is not None:
        cmap_sea = mcolors.ListedColormap([sea_color])
        cmap_sea.set_under(color="w", alpha=0)
        mask_sea = (z <= sea_level) * 1
        if mask_sea.any():
            axe.imshow(
                mask_sea,
                extent=im_extent,
                cmap=cmap_sea,
                vmin=0.5,
                interpolation="none",
            )

    return axe


def plot_imshow(
    x: np.ndarray,
    y: np.ndarray,
    data: np.ndarray,
    axe: matplotlib.axes._axes.Axes = None,
    figsize: tuple[float] = None,
    cmap: str = None,
    minval: int | float = None,
    maxval: int | float = None,
    vmin: float = None,
    vmax: float = None,
    alpha: float = 1.,
    minval_abs: float = None,
    cmap_intervals:  tuple[int] | tuple[float] = None,
    unique_values: bool = False,
    extend_cc: str = "max",
    plot_colorbar: bool = True,
    axecc: matplotlib.axes._axes.Axes = None,
    colorbar_kwargs: dict = None,
    aspect: str | float = None,
    ) -> matplotlib.axes._axes.Axes:
    """
    plt.imshow data with some pre-processing.

    Parameters
    ----------
    x : np.ndarray
        1D array of X coordinates.
    y : np.ndarray
        1D array of Y coordinates.
    data : np.ndarray
        2D array of data values.
    axe : matplotlib.axes._axes.Axes, optional
        Existing axes object to draw the plot on. The default is None.
    figsize : tuple[float], optional
        Figure size (width, height) if a new figure is created. The default is None.
    cmap : str, optional
        Name of the colormap to use. If None, use "hot_r" or "seismic". The default is None.
    minval : int or float, optional
        Minimum value to display. The default is None.
    maxval : int or float, optional
        Maximum value to display. The default is None.
    vmin : float, optional
        Alias for minval. The default is None.
    vmax : float, optional
        Alias for maxval. The default is None.
    alpha : float, optional
        Transparency for the data layer. The default is 1.
    minval_abs : float, optional
        If set, masks values with absolute magnitude below this threshold. The default is None.
    cmap_intervals : tuple of int or float, optional
        Explicit color intervals for categorical or segmented color mapping. The default is None.
    unique_values : bool, optional
        If True, use discrete colormap with ticks corresponding to unique values. The default is False.
    extend_cc : str, optional
        Behavior of colorbar outside bounds, can be "neither", "min", "max", "both". The default is "max".
    plot_colorbar : bool, optional
        Whether to add a colorbar. The default is True.
    axecc : matplotlib.axes._axes.Axes, optional
        Axes object for the colorbar. If None, colorbar is placed automatically. The default is None.
    colorbar_kwargs : dict, optional
        Additional keyword arguments passed to "figure.colorbar".
    aspect : str or float, optional
        Aspect ratio of the image. Can be "equal", "auto", or a numeric value.

    Returns
    -------
    axe : matplotlib.axes._axes.Axes
        Axes object with the image and optional colorbar.
    """
    if axe is None:
        _, axe = plt.subplots(1, 1, figsize=figsize, layout="constrained")

    f = copy.copy(data)

    # vmin and vmax are similar to minval and maxval
    # and supplent minval and maxval if used
    if vmin is not None:
        minval = vmin

    if vmax is not None:
        maxval = vmax

    # Remove values below and above minval and maxval, depending on whether
    # cmap_intervals are given with or without extend_cc
    if (cmap_intervals is not None) and not unique_values:
        if extend_cc in ["neither", "max"]:
            minval = cmap_intervals[0]
            f[f < minval] = np.nan
        elif extend_cc in ["neither", "min"]:
            maxval = cmap_intervals[-1]
            f[f > maxval] = np.nan
    else:
        if minval is not None:
            f[f < minval] = np.nan

    # Get min and max values
    if maxval is None:
        maxval = np.nanmax(f)
    if minval is None:
        minval = np.nanmin(f)

    if minval_abs:
        f[np.abs(f) <= minval_abs] = np.nan
    else:
        f[f == 0] = np.nan

    # Define colormap type
    if cmap is None:
        if maxval * minval >= 0:
            cmap = "hot_r"
        else:
            cmap = "seismic"

    norm = None
    if unique_values:
        if cmap_intervals is not None:
            values = cmap_intervals
        else:
            values = np.unique(f[~np.isnan(f)])
        n_values = len(values)
        if isinstance(cmap, list):
            color_map = mcolors.LinearSegmentedColormap.from_list('custom_cm', cmap, N=n_values)
        else:
            color_map = matplotlib.cm.get_cmap(cmap, n_values)
        bounds = np.zeros(n_values + 1)
        bounds[1:-1] = 0.5 * (values[1:] + values[:-1])
        bounds[0] = values[0] - 0.5
        bounds[-1] = values[-1] + 0.5
        norm = matplotlib.colors.BoundaryNorm(bounds, n_values)
        minval = None
        maxval = None
    else:
        if (maxval * minval >= 0) or np.isnan(maxval * minval):
            if isinstance(cmap, list):
                color_map = mcolors.LinearSegmentedColormap.from_list('custom_cm', cmap)
            else:
                color_map = matplotlib.colormaps[cmap]
        else:
            color_map = centered_map(cmap, minval, maxval)

    if cmap_intervals is not None and not unique_values:
        norm = matplotlib.colors.BoundaryNorm(
            cmap_intervals, 256, extend=extend_cc
        )
        maxval = None
        minval = None

    # get map_extent
    dx = x[1] - x[0]
    dy = y[1] - y[0]
    im_extent = [x[0] - dx / 2, x[-1] + dx / 2, y[0] - dy / 2, y[-1] + dy / 2]

    # Plot data

    fim = axe.imshow(
        f,
        extent=im_extent,
        cmap=color_map,
        vmin=minval,
        vmax=maxval,
        alpha=alpha,
        interpolation="none",
        norm=norm,
        zorder=4,
        aspect=aspect,
    )

    # Plot colorbar
    if plot_colorbar:
        colorbar_kwargs = {} if colorbar_kwargs is None else colorbar_kwargs
        if cmap_intervals is not None and extend_cc is not None:
            colorbar_kwargs["extend"] = extend_cc
        cb = axe.figure.colorbar(fim, cax=axecc, **colorbar_kwargs)
        if unique_values:
            cb.set_ticks(
                [(b0 + b1) / 2 for b0, b1 in zip(bounds[:-1], bounds[1:])],
                labels=values,
            )

    return axe


def plot_data_on_topo(
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray,
    data: np.ndarray,
    axe: matplotlib.axes._axes.Axes = None,
    figsize: tuple[float] = None,
    cmap: str = None,
    minval: int | float = None,
    maxval: int | float = None,
    vmin: float = None,
    vmax: float = None,
    minval_abs: float = None,
    cmap_intervals: tuple[int] | tuple[float] = None,
    unique_values: bool = False,
    extend_cc: str = "max",
    topo_kwargs: dict = None,
    alpha: float = 1.,
    plot_colorbar: bool = True,
    axecc: matplotlib.axes._axes.Axes = None,
    colorbar_kwargs: dict = None,
    mask: np.ndarray = None,
    alpha_mask: float = None,
    color_mask: str = "k",
    xlims: int | float = None,
    ylims: int | float = None,
    ) -> matplotlib.axes._axes.Axes:
    """
    Plot array data on topo.

    Parameters
    ----------
    x : np.ndarray
        1D array of X coordinates.
    y : np.ndarray
        1D array of Y coordinates.
    z : np.ndarray
        2D array of elevation values.
    data : np.ndarray
        2D array of data values.
    axe : matplotlib.axes._axes.Axes, optional
        Existing axes object to draw the plot on. The default is None.
    figsize : tuple[float], optional
        Figure size (width, height) if a new figure is created. The default is None.
    cmap : str, optional
        Name of the colormap to use. If None, use "hot_r" or "seismic". The default is None.
    minval : int or float, optional
        Minimum value to display. The default is None.
    maxval : int or float, optional
        Maximum value to display. The default is None.
    vmin : float, optional
        Alias for minval. The default is None.
    vmax : float, optional
        Alias for maxval. The default is None.
    minval_abs : float, optional
        If set, masks values with absolute magnitude below this threshold. The default is None.
    cmap_intervals : tuple of int or float, optional
        Explicit color intervals for categorical or segmented color mapping. The default is None.
    unique_values : bool, optional
        If True, use discrete colormap with ticks corresponding to unique values. The default is False.
    extend_cc : str, optional
        Behavior of colorbar outside bounds, can be "neither", "min", "max", "both". The default is "max".
    topo_kwargs : dict, optional
        Additional keyword arguments passed to "plot_topo". The default is None.
    alpha : float, optional
        Transparency for the data layer. The default is 1.
    plot_colorbar : bool, optional
        Whether to add a colorbar. The default is True.
    axecc : matplotlib.axes._axes.Axes, optional
        Axes object for the colorbar. If None, colorbar is placed automatically. The default is None.
    colorbar_kwargs : dict, optional
        Additional keyword arguments passed to "figure.colorbar".
    mask : np.ndarray, optional
        Binary 2D array of same shape as "data". Used to overlay a mask.
        Masked regions are displayed with "color_mask". The default is None.
    alpha_mask : float, optional
        Transparency of the mask overlay. The default is None (fully opaque).
    color_mask : str, optional
        Color used to draw the mask overlay. The default is "k".
    xlims : tuple of float, optional
        X-axis display limits (min, max). If None, auto-scaled. The default is None.
    ylims : tuple of float, optional
        Y-axis display limits (min, max). If None, auto-scaled. The default is None.

    Returns
    -------
    axe : matplotlib.axes._axes.Axes
        The axes object containing the final plot.
    """
    # Initialize figure properties
    dx = x[1] - x[0]
    dy = y[1] - y[0]
    im_extent = [x[0] - dx / 2, x[-1] + dx / 2, y[0] - dy / 2, y[-1] + dy / 2]
    if axe is None:
        fig, axe = plt.subplots(1, 1, figsize=figsize, layout="constrained")

    axe.set_ylabel("Y (m)")
    axe.set_xlabel("X (m)")
    axe.set_aspect("equal", adjustable="box")

    # Plot topo
    topo_kwargs = {} if topo_kwargs is None else topo_kwargs

    if z is not None:
        plot_topo(z, x, y, axe=axe, **topo_kwargs)

    # Plot mask
    if mask is not None:
        cmap_mask = mcolors.ListedColormap([color_mask])
        cmap_mask.set_under(color="w", alpha=0)
        axe.imshow(
            np.flip(mask, axis=0),
            extent=im_extent,
            cmap=cmap_mask,
            vmin=0.5,
            origin="lower",
            interpolation="none",
            # zorder=3,
            alpha=alpha_mask,
        )

    # Plot data
    plot_imshow(
        x,
        y,
        data,
        axe=axe,
        cmap=cmap,
        minval=minval,
        maxval=maxval,
        vmin=vmin,
        vmax=vmax,
        minval_abs=minval_abs,
        cmap_intervals=cmap_intervals,
        extend_cc=extend_cc,
        plot_colorbar=plot_colorbar,
        axecc=axecc,
        colorbar_kwargs=colorbar_kwargs,
        unique_values=unique_values,
        alpha=alpha,
    )

    # Adjust axes limits
    if xlims is not None:
        axe.set_xlim(xlims[0], xlims[1])

    if ylims is not None:
        axe.set_ylim(ylims[0], ylims[1])

    return axe


def plot_maps(
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray,
    data: np.ndarray,
    t: np.ndarray,
    file_name: str = None,
    folder_out: str = None,
    figsize: tuple[float] = None,
    dpi: int = None,
    fmt: str = "png",
    sup_plt_fn=None,
    sup_plt_fn_args=None,
    **kwargs
    ) -> None:
    """    
    Plot and save maps of simulations outputs at successive time steps

    Parameters
    ----------
    x : np.ndarray
        1D array of X coordinates.
    y : np.ndarray
        1D array of Y coordinates.
    z : np.ndarray
        2D array of elevation values.
    data : np.ndarray
        2D array of data values.
    t : np.ndarray
        List time steps, same length as the third dimension of "data".
    file_name : str, optional
        Base name for the output image files. The default is None.
    folder_out : str, optional
        Path to the output folder. If not provides, figures are not saved. The default is None.
    figsize : tuple[float], optional
        Size of each figure (width, height). The default is None.
    dpi : int, optional
        Resolution for saved figures. Only used if "folder_out" is set. The default is None.
    fmt : str, optional
        File format for saving figures. Default is "png".
    sup_plt_fn : callable, optional
        A custom function to apply additional plotting on the axes. The default is None.
    sup_plt_fn_args : dict, optional
        Arguments to pass to "sup_plt_fn". The default is None.
    **kwargs : dict
        Additional keyword arguments passed to "plot_data_on_topo".
    """
    nfigs = len(t)
    if nfigs != data.shape[2]:
        raise ValueError(
            "length of t must be similar to the last dimension of data"
        )
    if folder_out is not None:
        file_path = os.path.join(folder_out, file_name + "_{:04d}." + fmt)
    title_fmt = "t = {:.2f} s"

    for i in range(nfigs):
        axe = plot_data_on_topo(
            x, y, z, data[:, :, i], axe=None, figsize=figsize, **kwargs
        )
        axe.set_title(title_fmt.format(t[i]))
        if sup_plt_fn is not None:
            if sup_plt_fn_args is None:
                sup_plt_fn_args = dict()
            sup_plt_fn(axe, **sup_plt_fn_args)
        # axe.figure.tight_layout(pad=0.1)
        if folder_out is not None:
            axe.figure.savefig(
                file_path.format(i),
                dpi=dpi,
                bbox_inches="tight",
                pad_inches=0.05,
            )